/*
 * Copyright(c) 1997-2000 - All Rights Reserved
 *
 * See the COPYRIGHT file.
 */

#ifndef _option_h_
#define _option_h_

void	opt_init();
void	opt_cfg();

void	opt_free(struct custom_ *);
void	sic_logfile(struct log_ *);

char *	get_format(unsigned int, char *);
char *	get_nformat(unsigned int, char *);
int	get_option(unsigned int, char *);
int	get_flag(unsigned int, char *, unsigned int);
int	get_keyword(char *, char **, char *, unsigned int *);	
void	get_ignore(char *, char *, unsigned int *);
int	get_rewrite(int, char *, u_char *, unsigned int *, char *);

void	customize(unsigned int, char *, int, char **, unsigned int *);

#define	O_CLEAR		0x00001
#define	O_GET		0x00002
#define	O_SET		0x00004

#define	O_ON		0x00010
#define	O_OFF		0x00020
#define	O_MASK		0x00040
#define	O_KEYWORD	0x00080
#define	O_LOGFILE	0x00100
#define	O_SWITCH	0x00200
#define	O_REWRITE	0x00400

#define	O_WINDOW	0x01000
#define	O_SERVER	0x02000
#define	O_TOPLVL	0x04000
#define	O_PROTO		0x08000
#define	O_DEFAULT	0x10000

#define	O_ACTIONMASK	0x00007
#define	O_TYPEMASK	0x007F0
#define	O_TARGETMASK	0x1F000

#endif
