/*
 * Copyright(c) 1997-2000 - All Rights Reserved
 *
 * See the COPYRIGHT file.
 */

struct channel_	*get_channel(char *);

void		sic_log(unsigned int, time_t, char *, char *, char *, char *,
			char, char *, u_char *);
void		vsic_log(unsigned int, char *, char, char *, char *,
			 char, char *, ...);
void		vsic_slog(unsigned int, char *, ...);
#define		sic_slog(x, y)	sic_log(x, 0, NULL, NULL, NULL, NULL, -1, y, NULL)
void		sic_dlog(int);
void		sic_clog(int);

void		display_status();

void		sic_newwin();
char		sic_swin(char);
void		sic_wkill();
void		sic_wlist();
void		sic_wch(signed char);
void		sic_wdcc();
void		sic_wchannel(char *);

void		sic_redowin(int);
void		sic_chgwin(int);
void		sic_wptoggle();
void		sic_scroll(int);
void		sic_scroll2(int, char *);

int		map_c2w(char *);
void		chg_channel(char *);
void		new_channel(char *);
void		join_channel(char *, int);
void		part_channel(char *, char *, int);
void		del_channel(char *);
void		rejoin_channel(char *);
void		collect_channel();
void		default_channel(char *);
char *		get_query();
void		select_query(char *);
void		select_active(char *, int);

void		set_chop(char *, int);
void		add_member(char *, char *, char);
void		del_member(char *, char *);
void		chg_member(char *, char *);
int		is_member(char *, char *);
