/*
 * Copyright(c) 1998-2000 - All Rights Reserved
 *
 * See the COPYRIGHT file.
 */

#ifndef _counter_h_
#define _counter_h_

void	counter_init();
# if defined(HAVE_REGEXP)
void	counter_expire(int);
int	counter_add(void *,char *,u_char *,regmatch_t *,unsigned int *,char *);
# endif

#endif
