/*
 * Copyright(c) 1997-2000 - All Rights Reserved
 *
 * See the COPYRIGHT file.
 */

#ifndef _utils_h_
#define _utils_h_

extern	char *	strcasestr(const char *, const char *);
extern	int	rmatch(const char *, const char *);
extern	char *	my_cftime(char *, const int len, const char *, const time_t);
extern	char *	my_strftime(char *, const int len, const char *);
extern	char *	sic_tdiff(time_t, char);
# if !defined(HAVE_STRERROR)
extern	char *	strerror(int);
# endif
# if defined(HAVE_GETADDRINFO)
extern	char *	aitoip(struct addrinfo *);
# endif

#endif
